#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <sstream>
#include <vector>
#ifdef __APPLE__
#include <GL/glew.h>
#else
#include <OpenGL/glew.h>
#endif
#include <GLUT/glut.h>

#pragma clang diagnostic ignored "-Wdeprecated-declarations"

using namespace std;

GLuint* vaos;
GLuint* progIds;

void display() {
	glClear(GL_COLOR_BUFFER_BIT);
	for (int i = 0; i < 2; i++) {
		glUseProgram(progIds[i]);
		glBindVertexArray(vaos[i]);
		glDrawArrays(GL_TRIANGLES, 0, 3);
	}
	glutSwapBuffers();
}

void attachShader(GLuint progId, const char* filePath, GLenum type) {
	GLuint shader = glCreateShader(type);

	if (!shader) {
		fprintf(stderr, "error creating shader\n");
		exit(1);
	}

	ifstream file(filePath);
	stringstream buffer;
	buffer << file.rdbuf();
	const char* contents = buffer.str().c_str();

	glShaderSource(shader, 1, (const GLchar**)&contents, NULL);
	glCompileShader(shader);
	GLint success;
	glGetShaderiv(shader, GL_COMPILE_STATUS, &success);
	if (!success) {
		GLchar log[1024];
		glGetShaderInfoLog(shader, 1024, NULL, log);
		fprintf(stderr, "error: %s\n", log);
		exit(1);
	}
	glAttachShader(progId, shader);
}

GLuint compileShaders(char* vertexShader, char* fragmentShader) {
	GLuint progId = glCreateProgram();

	attachShader(progId, vertexShader, GL_VERTEX_SHADER);
	attachShader(progId, fragmentShader, GL_FRAGMENT_SHADER);

	glLinkProgram(progId);
	GLint success = 0;
	glGetProgramiv(progId, GL_LINK_STATUS, &success);
	if (!success) {
		GLchar log[1024];
		glGetProgramInfoLog(progId, sizeof(log), NULL, log);
		fprintf(stderr, "error linking: %s\n", log);
		exit(1);
	}

	return progId;
}

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

GLuint setupBuffers(GLfloat* vertices, GLuint progId) {

	GLfloat colors[] = {
		0, 1, 0, 1,
		1, 0, 0, 1,
		0, 0, 1, 1
	};

	GLuint numVerts = 3;

	GLuint vbo;
	glGenBuffers(1, &vbo);

	GLuint vao;
	glGenVertexArrays(1, &vao);

	GLuint posId = glGetAttribLocation(progId, "vPosition");
	GLuint colorId = glGetAttribLocation(progId, "vColor");

	GLuint vertsLen = numVerts * 3 * sizeof(GLfloat);
	GLuint colorsLen = numVerts * 4 * sizeof(GLfloat);

	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, vertsLen + colorsLen, NULL, GL_STATIC_DRAW);

	glBufferSubData(GL_ARRAY_BUFFER, 0, vertsLen, vertices);
	glBufferSubData(GL_ARRAY_BUFFER, vertsLen, colorsLen, colors);
	
	glBindVertexArray(vao);

	glEnableVertexAttribArray(posId);
	glEnableVertexAttribArray(colorId);
	
	glVertexAttribPointer(posId, 3, GL_FLOAT, GL_FALSE, 0, 0);
	glVertexAttribPointer(colorId, 4, GL_FLOAT, GL_FALSE, 0, BUFFER_OFFSET(numVerts * 3 * sizeof(GLfloat)));

	return vao;
}

void validateProgram(GLuint progId) {
	glValidateProgram(progId);
	
	GLint success;
	glGetProgramiv(progId, GL_VALIDATE_STATUS, &success);
	if (!success) {
		GLchar log[1024];
		glGetProgramInfoLog(progId, sizeof(log), NULL, log);
		fprintf(stderr, "error: %s\n", log);
		exit(1);
	}
}

void init() {
	GLfloat vertices[2][9] = {
		{
			0.0f, -1.0f, 0.0f,
			1.0f, -1.0f, 0.0f,
			0.5f, 1.0f, 0.0f
		},

		{
			-1.0f, -1.0f, 0.0f,
			0.0f, -1.0f, 0.0f,
			-0.5f, 1.0f, 0.0f
		}
	};

	vaos = new GLuint[2];
	progIds = new GLuint[2];

	GLuint progId1 = compileShaders((char*)"vertex.glsl", (char*)"fragment.glsl");
	vaos[0] = setupBuffers(vertices[0], progId1);
	progIds[0] = progId1;
	validateProgram(progId1);
	
	GLuint progId2 = compileShaders((char*)"vertex.glsl", (char*)"yellow.glsl");
	vaos[1] = setupBuffers(vertices[1], progId2);
	progIds[1] = progId2;
	validateProgram(progId2);
}

int main(int argc, char** argv) {
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB|GLUT_3_2_CORE_PROFILE);
	glutInitWindowSize(800, 600);
	glutCreateWindow("Hello Triangle");
	glutDisplayFunc(display);

	glewInit();
	
	init();

	glutMainLoop();

	return 0;
}

